pub mod model;
pub use model::historic::Historic;
pub use model::reference::*;
pub use model::valid::Validate;
pub use model::Model;
pub use model::Sort;
pub use mongod_derive as derive;
pub use mongodb;

#[cfg(feature = "cache")]
pub mod cache;
#[cfg(feature = "cache")]
pub use cache::ModelCache;

use tokio::sync::OnceCell;

#[cfg(feature = "cache")]
use std::sync::RwLock;

pub static MONGO_CLIENT: OnceCell<mongodb::Client> = OnceCell::const_new();

#[cfg(feature = "cache")]
pub static DB_CACHE: OnceCell<RwLock<ModelCache>> = OnceCell::const_new();

#[cfg(feature = "cache")]
#[macro_export]
macro_rules! cache_read {
    () => {
        if let Some(cache) = $crate::DB_CACHE.get() {
            cache.read().unwrap()
        } else {
            let cache = std::sync::RwLock::new($crate::ModelCache::new());
            $crate::DB_CACHE.set(cache).unwrap();
            $crate::DB_CACHE.get().unwrap().read().unwrap()
        }
    };
}

#[cfg(feature = "cache")]
#[macro_export]
macro_rules! cache_write {
    () => {
        if let Some(cache) = $crate::DB_CACHE.get() {
            cache.write().unwrap()
        } else {
            let cache = std::sync::RwLock::new($crate::ModelCache::new());
            $crate::DB_CACHE.set(cache).unwrap();
            $crate::DB_CACHE.get().unwrap().write().unwrap()
        }
    };
}

/// Get a `MongoDB` Client from the environment
#[macro_export]
macro_rules! get_mongo {
    () => {
        if let Some(client) = $crate::MONGO_CLIENT.get() {
            client
        } else {
            let client = $crate::mongodb::Client::with_uri_str(&std::env::var("DB_URI").unwrap())
                .await
                .unwrap();
            $crate::MONGO_CLIENT.set(client).unwrap();
            $crate::MONGO_CLIENT.get().unwrap()
        }
    };
}

/// Get a database collection
#[macro_export]
macro_rules! col {
    ($db:expr, $col:expr) => {
        $db.database(&std::env::var("DB").unwrap())
            .collection::<$crate::mongodb::bson::Document>($col)
    };
}

/// Collect database results into a `Vec<_>`
#[macro_export]
macro_rules! collect_results {
    ($res:expr) => {{
        use futures::stream::TryStreamExt;
        let mut ret = vec![];

        while let Some(doc) = $res.try_next().await.unwrap() {
            ret.push(doc);
        }

        ret
    }};
}

/// `MongoDB` filter for the `_id` field.
#[macro_export]
macro_rules! id_of {
    ($id:expr) => {
        $crate::mongodb::bson::doc! { "_id": $id}
    };
}

/// A trait to generate a Model API representation in JSON format.
pub trait ToAPI: Sized {
    /// Generate public API JSON
    fn api(&self) -> impl std::future::Future<Output = serde_json::Value>;
}

/// Converts a slice of items implementing the `ToAPI` trait into a `Vec` of JSON values.
pub async fn vec_to_api(items: &[impl ToAPI]) -> Vec<serde_json::Value> {
    let mut ret = Vec::with_capacity(items.len());

    for e in items {
        ret.push(e.api().await);
    }

    ret
}
