use mongodb::{
    bson::doc,
    options::{FindOneOptions, FindOptions},
    results::{DeleteResult, InsertOneResult},
};
use reference::Referencable;
use serde::de::DeserializeOwned;
use serde_json::{Map, Value};
use valid::Validate;

#[cfg(feature = "cache")]
use crate::{cache_read, cache_write};

use crate::{col, collect_results, get_mongo, id_of};

pub mod historic;
pub mod reference;
pub mod update;
pub mod valid;

/// Error type when updating a model
#[derive(Debug)]
pub enum UpdateError {
    /// Provided data was no object
    NoObject,
    /// Database related error
    Database(mongodb::error::Error),
    /// Validation failed
    Validation(String),
}

pub trait Model:
    Sized + Referencable + Validate + serde::Serialize + for<'a> serde::Deserialize<'a>
{
    type Partial: DeserializeOwned;
    type ChangeBuilder;

    /// Insert the `Model` into the database
    fn insert(
        &self,
    ) -> impl std::future::Future<Output = Result<InsertOneResult, mongodb::error::Error>> + Send
    where
        Self: Sync,
    {
        #[cfg(feature = "cache")]
        {
            cache_write!().insert(
                Self::collection_name(),
                self.id(),
                mongodb::bson::to_document(self).unwrap(),
            );
        }

        async {
            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());
            collection
                .insert_one(mongodb::bson::to_document(self).unwrap(), None)
                .await
        }
    }

    /// Insert or update an existing `Model`
    fn insert_overwrite(
        &self,
    ) -> impl std::future::Future<Output = Result<(), mongodb::error::Error>> + Send
    where
        Self: Sync,
    {
        async {
            if self.insert().await.is_err() {
                let db = get_mongo!();
                let collection = col!(db, Self::collection_name());

                collection
                    .update_one(
                        id_of!(self.id()),
                        mongodb::bson::doc! {"$set": mongodb::bson::to_document(self).unwrap() },
                        None,
                    )
                    .await?;
            }

            Ok(())
        }
    }

    /// Remove a `Model` from the database.
    #[must_use]
    fn remove(
        id: &str,
    ) -> impl std::future::Future<Output = Result<DeleteResult, mongodb::error::Error>> + Send {
        async move {
            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());

            #[cfg(feature = "cache")]
            {
                cache_write!().invalidate(Self::collection_name(), id);
            }

            collection.delete_one(id_of!(id), None).await
        }
    }

    /// Remove a `Model` from the database.
    ///
    /// This is a convenience function to let you call `remove()` on a `Model` you have at hand.
    fn delete(
        &self,
    ) -> impl std::future::Future<Output = Result<DeleteResult, mongodb::error::Error>> + Send
    where
        Self: Sync,
    {
        async { Self::remove(self.id()).await }
    }

    /// Get a `Model` by id from the database
    #[must_use]
    fn get(id: &str) -> impl std::future::Future<Output = Option<Self>> {
        async move {
            #[cfg(feature = "cache")]
            {
                if let Some(cached) = cache_read!().get(Self::collection_name(), id) {
                    return mongodb::bson::from_document(cached.clone()).ok();
                }
            }

            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());
            let doc = collection.find_one(id_of!(id), None).await.ok()??;
            let res = mongodb::bson::from_document(doc.clone()).ok();

            #[cfg(feature = "cache")]
            {
                if res.is_some() {
                    cache_write!().insert(Self::collection_name(), id, doc);
                }
            }

            res
        }
    }

    /// Get a partial `Model` by id from the database with only some fields retrieved.
    #[must_use]
    fn get_partial(
        id: &str,
        mut part: serde_json::Value,
    ) -> impl std::future::Future<Output = Option<Self::Partial>> {
        async move {
            #[cfg(feature = "cache")]
            {
                // todo : implement
            }

            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());

            part.as_object_mut()?.insert("_id".into(), 1.into());

            let doc = collection
                .find_one(
                    id_of!(id),
                    Some(
                        FindOneOptions::builder()
                            .projection(Some(mongodb::bson::to_document(&part).unwrap()))
                            .build(),
                    ),
                )
                .await
                .ok()??;
            mongodb::bson::from_document(doc).ok()
        }
    }

    /// Get a `Model` by using a filter from the database
    #[must_use]
    fn find_one_partial(
        filter: mongodb::bson::Document,
        mut part: serde_json::Value,
    ) -> impl std::future::Future<Output = Option<Self::Partial>> {
        async move {
            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());

            part.as_object_mut()?.insert("_id".into(), 1.into());

            let doc = collection
                .find_one(
                    filter,
                    Some(
                        FindOneOptions::builder()
                            .projection(Some(mongodb::bson::to_document(&part).unwrap()))
                            .build(),
                    ),
                )
                .await
                .ok()??;
            mongodb::bson::from_document(doc).ok()
        }
    }

    /// Get a partial `Model` by using a filter from the database
    #[must_use]
    fn find_one(
        filter: mongodb::bson::Document,
    ) -> impl std::future::Future<Output = Option<Self>> {
        async move {
            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());
            let doc = collection.find_one(filter, None).await.ok()??;
            mongodb::bson::from_document(doc).ok()
        }
    }

    /// Get all `Model`s from the database
    #[must_use]
    fn find_all() -> impl std::future::Future<Output = Option<Vec<Self>>> {
        Self::find(doc! {}, None, None)
    }

    /// Get all `Model`s partial from the database
    #[must_use]
    fn find_all_partial(
        part: serde_json::Value,
    ) -> impl std::future::Future<Output = Option<Vec<Self::Partial>>> {
        Self::find_partial(doc! {}, part, None, None)
    }

    /// Get multiple `Model`s by using a filter from the database. Pass a `limit` parameter to limit the amount of `Model`s returned.
    #[must_use]
    fn find(
        filter: mongodb::bson::Document,
        limit: Option<i64>,
        sort: Option<mongodb::bson::Document>,
    ) -> impl std::future::Future<Output = Option<Vec<Self>>> {
        async move {
            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());

            let options = FindOptions::builder().limit(limit).sort(sort).build();

            let mut results = collection.find(filter, options).await.ok()?;
            let docs = collect_results!(results);
            docs.into_iter()
                .map(|x| mongodb::bson::from_document(x).unwrap())
                .collect()
        }
    }

    /// Get unique values of a Models field
    fn unique<T: DeserializeOwned>(
        filter: mongodb::bson::Document,
        field: &str,
    ) -> impl std::future::Future<Output = Vec<T>> {
        async move {
            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());
            let res = collection.distinct(field, filter, None).await.unwrap();
            res.into_iter()
                .filter_map(|x| mongodb::bson::from_bson(x).ok())
                .collect()
        }
    }

    /// Get multiple partial `Model`s by using a filter from the database. Pass a `limit` parameter to limit the amount of `Model`s returned.
    #[must_use]
    fn find_partial(
        filter: mongodb::bson::Document,
        mut part: serde_json::Value,
        limit: Option<i64>,
        sort: Option<mongodb::bson::Document>,
    ) -> impl std::future::Future<Output = Option<Vec<Self::Partial>>> {
        async move {
            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());

            part.as_object_mut()?.insert("_id".into(), 1.into());

            let options = FindOptions::builder()
                .limit(limit)
                .sort(sort)
                .projection(Some(mongodb::bson::to_document(&part).unwrap()))
                .build();

            let mut results = collection.find(filter, Some(options)).await.ok()?;
            let docs = collect_results!(results);
            docs.into_iter()
                .map(|x| mongodb::bson::from_document(x).unwrap())
                .collect()
        }
    }

    fn change_builder(self) -> Self::ChangeBuilder;

    /// Update values of `Model` using a builder pattern
    fn change(self) -> Self::ChangeBuilder {
        #[cfg(feature = "cache")]
        {
            cache_write!().invalidate(Self::collection_name(), self.id());
        }

        self.change_builder()
    }

    /// Update values of `Model` into database
    fn update(
        &mut self,
        data: &serde_json::Value,
    ) -> impl std::future::Future<Output = Result<(), UpdateError>> {
        async {
            // get db collection
            let db = get_mongo!();
            let collection = col!(db, Self::collection_name());
            let mut update = mongodb::bson::Document::new();

            if let Some(obj) = data.as_object() {
                // run model specific update
                self.update_values(obj, &mut update).await;

                // validate and update
                if let Err(msg) = self.validate().await {
                    return Err(UpdateError::Validation(msg));
                }
                collection
                    .update_one(
                        id_of!(self.id()),
                        mongodb::bson::doc! {"$set": update },
                        None,
                    )
                    .await
                    .map_err(UpdateError::Database)?;

                #[cfg(feature = "cache")]
                {
                    cache_write!().invalidate(Self::collection_name(), self.id());
                }

                return Ok(());
            }

            Err(UpdateError::NoObject)
        }
    }

    /// Update the `Model` based on the provided JSON Object.
    ///
    /// This function should be implemented for a `Model`. It should update all keys from the JSON Object.
    /// For every updated value in the JSON Object the `Model`s fields should be updated and the value should be put in the `update` `Document`
    ///
    /// To avoid making mistakes in the Update logic use the macros from `update.rs`
    fn update_values(
        &mut self,
        obj: &Map<String, Value>,
        update: &mut mongodb::bson::Document,
    ) -> impl std::future::Future<Output = ()> + Send;
}

/// Sorting Order
///
/// # Example
/// ```ignore
/// let m = MyModel::find(doc! {}, None, Some(doc! {
///     "sort_key": Sort::Ascending
/// })).await.unwrap();
/// ```
#[derive(Debug)]
pub enum Sort {
    /// Sort in ascending order
    Ascending,
    /// Sort in descending order
    Descending,
}

impl Into<mongodb::bson::Bson> for Sort {
    fn into(self) -> mongodb::bson::Bson {
        match self {
            Sort::Ascending => mongodb::bson::bson!(1),
            Sort::Descending => mongodb::bson::bson!(-1),
        }
    }
}
